import io
import platform
import subprocess
import base64

def get_clipboard_content() -> tuple[str | bytes, str | None] | None:
    """Get content from clipboard, handling both text and images in native and WSL environments.
    Returns:
        tuple: (content, mime_type) where:
            - content is either string (for text) or bytes (for image)
            - mime_type is None for text or mime type string for images
        None: if clipboard is empty or error occurs
    """
    system = platform.system()
    is_wsl = 'microsoft-standard' in platform.uname().release.lower()
    
    if is_wsl or system == 'Windows':
        try:
            # Try to get image or text from Windows clipboard
            ps_script = '''
            Add-Type -AssemblyName System.Windows.Forms
            if ([Windows.Forms.Clipboard]::ContainsImage()) {
                $image = [Windows.Forms.Clipboard]::GetImage()
                $ms = New-Object System.IO.MemoryStream
                $image.Save($ms, [System.Drawing.Imaging.ImageFormat]::Png)
                Write-Output "IMAGE:"
                [Convert]::ToBase64String($ms.ToArray())
            } elseif ([Windows.Forms.Clipboard]::ContainsText()) {
                Write-Output "TEXT:"
                [Windows.Forms.Clipboard]::GetText()
            }
            '''
            # Execute PowerShell script
            powershell_cmd = 'powershell.exe' if is_wsl else 'powershell'
            result = subprocess.run(
                [powershell_cmd, '-Command', ps_script],
                capture_output=True,
                text=True
            )
            if result.stdout.strip():
                lines = result.stdout.strip().split('\n', 1)
                if len(lines) == 2:
                    content_type, content = lines
                    if content_type.strip() == "IMAGE:":
                        # Convert base64 back to bytes
                        img_bytes = base64.b64decode(content.strip())
                        return img_bytes, 'image/png'
                    elif content_type.strip() == "TEXT:":
                        return content.strip(), None
        except Exception as e:
            print(f"Error accessing Windows clipboard: {e}")
            return None
    elif system == 'Darwin':  # macOS
        try:
            # Try image first (using pngpaste if available)
            result = subprocess.run(['pngpaste', '-'], capture_output=True)
            if result.returncode == 0 and result.stdout:
                return result.stdout, 'image/png'
            
            # If no image, try text
            result = subprocess.run(['pbpaste'], capture_output=True, text=True)
            if result.stdout:
                return result.stdout.strip(), None
        except FileNotFoundError:
            # pngpaste not installed
            print("Error: pngpaste not installed. Install it with 'brew install pngpaste' for image clipboard support")
            try:
                result = subprocess.run(['pbpaste'], capture_output=True, text=True)
                if result.stdout:
                    return result.stdout.strip(), None
            except:
                pass
        except Exception as e:
            print(f"Error accessing macOS clipboard: {e}")
            raise e
    elif system == 'Linux':  # Linux
        try:
            # Try image first
            result = subprocess.run(
                ['xclip', '-selection', 'clipboard', '-t', 'image/png', '-o'],
                capture_output=True
            )
            if result.returncode == 0 and result.stdout:
                return result.stdout, 'image/png'
            
            # If no image, try text
            result = subprocess.run(
                ['xclip', '-selection', 'clipboard', '-o'],
                capture_output=True,
                text=True
            )
            if result.stdout:
                return result.stdout.strip(), None
        except Exception as e:
            print(f"Error accessing Linux clipboard: {e}")
            raise e

    raise Exception("Clipboard is empty")


import json
from langchain_core.messages import BaseMessage, AIMessage, AIMessageChunk, ToolMessage
from rich.console import Console, ConsoleDimensions
from rich.live import Live
from rich.markdown import Markdown
from rich.prompt import Confirm

class OutputHandler:
    def __init__(self, text_only: bool = False, only_last_message: bool = False):
        self.console = Console()
        self.text_only = text_only
        self.only_last_message = only_last_message
        self.last_message = ""
        if self.text_only:
            self.md = ""
        else:
            self.md = "Thinking...\n"
        self._live = None

    def start(self):
        if not self.text_only:
            self._live = Live(
                Markdown(self.md), 
                vertical_overflow="visible", 
                screen=True,
                console=self.console
            )
            self._live.start()

    def update(self, chunk: any):
        self.md = self._parse_chunk(chunk, self.md)
        if(self.only_last_message and self.text_only):
            # when only_last_message, we print in finish()
            return
        if self.text_only:
            self.console.print(self._parse_chunk(chunk), end="")
        else:
            if self.md.startswith("Thinking...") and not self.md.strip("Thinking...").isspace():
                self.md = self.md.strip("Thinking...").strip()
            partial_md = self._truncate_md_to_fit(self.md, self.console.size)
            self._live.update(Markdown(partial_md), refresh=True)

    def update_error(self, error: Exception):
        import traceback
        error = f"Error: {error}\n\nStack trace:\n```\n{traceback.format_exc()}```"
        self.md += error
        if(self.only_last_message):
            self.console.print(error)
            return
        if self.text_only:
            self.console.print_exception()
        else:
            partial_md = self._truncate_md_to_fit(self.md, self.console.size)
            self._live.update(Markdown(partial_md), refresh=True)

    def stop(self):
        if not self.text_only and self._live:
            self._live.stop()

    def confirm_tool_call(self, config: dict, chunk: any) -> bool:
        if not self._is_tool_call_requested(chunk, config):
            return True

        self.stop()
        is_confirmed = self._ask_tool_call_confirmation()
        if not is_confirmed:
            self.md += "# Tool call denied"
            return False
            
        if not self.text_only:
            self.start()
        return True

    def finish(self):
        self.stop()
        to_print = self.last_message if self.only_last_message else Markdown(self.md)
        if not self.text_only and not self.only_last_message:
            self.console.clear()
            self.console.print(Markdown(self.md))
        if self.only_last_message:
            self.console.print(to_print)

    def _parse_chunk(self, chunk: any, md: str = "") -> str:
        """
        Parse the chunk of agent response.
        It will stream the response as it is received.
        """
        # If this is a message chunk
        if isinstance(chunk, tuple) and chunk[0] == "messages":
            message_chunk = chunk[1][0]  # Get the message content
            if isinstance(message_chunk, AIMessageChunk):
                content = message_chunk.content
                self.last_message += content
                if isinstance(content, str):
                    md += content
                elif isinstance(content, list) and len(content) > 0 and isinstance(content[0], dict) and "text" in content[0]:
                    md += content[0]["text"]
        # If this is a final value
        elif isinstance(chunk, dict) and "messages" in chunk:
            # Print a newline after the complete message
            md += "\n"
            self.last_message = ""
        elif isinstance(chunk, tuple) and chunk[0] == "values":
            message: BaseMessage = chunk[1]['messages'][-1]
            if isinstance(message, AIMessage) and message.tool_calls:
                md += "\n\n### Tool Calls:"
                for tc in message.tool_calls:
                    lines = [
                        f"  {tc.get('name', 'Tool')}",
                    ]
                    if tc.get("error"):
                        lines.append(f"```")
                        lines.append(f"Error: {tc.get('error')}")
                        lines.append("```")

                    lines.append("Args:")
                    lines.append("```")
                    args = tc.get("args")
                    if isinstance(args, str):
                        lines.append(f"{args}")
                    elif isinstance(args, dict):
                        for arg, value in args.items():
                            lines.append(f"{arg}: {value}")
                    lines.append("```\n")
                    md += "\n".join(lines)
                self.last_message = ""
            elif isinstance(message, ToolMessage) and message.status != "success":
                md += "Failed call with error:"
                md += f"\n\n{message.content}"
            md += "\n"
        return md

    def _truncate_md_to_fit(self, md: str, dimensions: ConsoleDimensions) -> str:
        """
        Truncate the markdown to fit the console size, with few line safety margin.
        """
        lines = md.splitlines()
        max_lines = dimensions.height - 3  # Safety margin
        fitted_lines = []
        current_height = 0
        code_block_count = 0

        for line in reversed(lines):
            # Calculate wrapped line height, rounding up for safety
            line_height = 1 + len(line) // dimensions.width

            if current_height + line_height > max_lines:
                # If we're breaking in the middle of code blocks, add closing ```
                if code_block_count % 2 == 1:
                    fitted_lines.insert(0, "```")
                break

            fitted_lines.insert(0, line)
            current_height += line_height

            # Track code block markers
            if line.strip() == "```":
                code_block_count += 1

        return '\n'.join(fitted_lines) if fitted_lines else ''

    def _is_tool_call_requested(self, chunk: any, config: dict) -> bool:
        """
        Check if the chunk contains a tool call request and requires confirmation.
        """
        if isinstance(chunk, tuple) and chunk[0] == "values":
            if len(chunk) > 1 and isinstance(chunk[1], dict) and "messages" in chunk[1]:
                message = chunk[1]['messages'][-1]
                if isinstance(message, AIMessage) and message.tool_calls:
                    for tc in message.tool_calls:
                        if tc.get("name") in config["tools_requires_confirmation"]:
                            return True
        return False

    def _ask_tool_call_confirmation(self) -> bool:
        """
        Ask the user for confirmation to run a tool call.
        """
        self.console.set_alt_screen(True)
        self.console.print(Markdown(self.md))
        self.console.print(f"\n\n")
        is_tool_call_confirmed = Confirm.ask(f"Confirm tool call?", console=self.console)
        self.console.set_alt_screen(False)
        if not is_tool_call_confirmed:
            return False
        return True


"""SQLite-based store implementation with vector search capabilities.

This store provides persistent storage using SQLite with optional vector search functionality.
It implements the BaseStore interface from langgraph.
"""

from datetime import datetime, timezone
import json
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated
import uuid

import aiosqlite
from langchain_core.embeddings import Embeddings
from langgraph.prebuilt import InjectedStore
from langgraph.store.base import BaseStore
from langchain_core.runnables import RunnableConfig
from langchain_core.tools import tool

from langgraph.store.base import (
    BaseStore,
    GetOp,
    IndexConfig,
    Item,
    ListNamespacesOp,
    MatchCondition,
    Op,
    PutOp,
    Result,
    SearchItem,
    SearchOp,
    ensure_embeddings,
    get_text_at_path,
    tokenize_path,
)

logger = logging.getLogger(__name__)
        

@tool
async def save_memory(memories: List[str], *, config: RunnableConfig, store: Annotated[BaseStore, InjectedStore()]) -> str:
    '''Save the given memory for the current user. Do not save duplicate memories.'''
    user_id = config.get("configurable", {}).get("user_id")
    namespace = ("memories", user_id)
    for memory in memories:
        id = uuid.uuid4().hex
        await store.aput(namespace, f"memory_{id}", {"data": memory})
    return f"Saved memories: {memories}"

async def get_memories(store: BaseStore, user_id: str = "myself", query: str = None) -> List[str]:
    namespace = ("memories", user_id)
    memories = [m.value["data"] for m in await store.asearch(namespace, query=query)]
    return memories

class SqliteStore(BaseStore):
    """SQLite-based store with optional vector search.

    This store provides persistent storage using SQLite with optional vector search functionality.
    Data is stored in two tables:
    - items: Stores the actual key-value pairs with their metadata
    - vectors: Stores vector embeddings for semantic search

    Args:
        db_path (Union[str, Path]): Path to the SQLite database file
        index (Optional[IndexConfig]): Configuration for vector search functionality
    """

    def __init__(
        self, db_path: Union[str, Path], *, index: Optional[IndexConfig] = None
    ) -> None:
        self.db_path = Path(db_path)
        self.db_path.parent.mkdir(parents=True, exist_ok=True)
        self.index_config = index
        if self.index_config:
            self.index_config = self.index_config.copy()
            self.embeddings: Optional[Embeddings] = ensure_embeddings(
                self.index_config.get("embed"),
            )
            self.index_config["__tokenized_fields"] = [
                (p, tokenize_path(p)) if p != "$" else (p, p)
                for p in (self.index_config.get("fields") or ["$"])
            ]
        else:
            self.index_config = None
            self.embeddings = None

    async def _init_db(self, db: aiosqlite.Connection) -> None:
        """Initialize database schema.

        Args:
            db (aiosqlite.Connection): Database connection
        """
        await db.execute("""
            CREATE TABLE IF NOT EXISTS items (
                namespace TEXT,
                key TEXT,
                value TEXT,
                created_at TEXT,
                updated_at TEXT,
                PRIMARY KEY (namespace, key)
            )
        """)
        if self.index_config:
            await db.execute("""
                CREATE TABLE IF NOT EXISTS vectors (
                    namespace TEXT,
                    key TEXT,
                    path TEXT,
                    vector BLOB,
                    PRIMARY KEY (namespace, key, path),
                    FOREIGN KEY (namespace, key) REFERENCES items (namespace, key)
                        ON DELETE CASCADE
                )
            """)
        await db.commit()

    def batch(self, ops: List[Op]) -> List[Result]:
        """Execute a batch of operations synchronously.

        Args:
            ops (List[Op]): List of operations to execute

        Returns:
            List[Result]: Results of the operations
        """
        raise NotImplementedError

    async def abatch(self, ops: List[Op]) -> List[Result]:
        """Execute a batch of operations asynchronously.

        Args:
            ops (List[Op]): List of operations to execute

        Returns:
            List[Result]: Results of the operations
        """
        async with aiosqlite.connect(self.db_path) as db:
            await self._init_db(db)
            results: List[Result] = []
            put_ops: Dict[Tuple[Tuple[str, ...], str], PutOp] = {}
            search_ops: Dict[int, Tuple[SearchOp, List[Tuple[Item, List[List[float]]]]]] = {}

            for i, op in enumerate(ops):
                if isinstance(op, GetOp):
                    item = await self._get_item(db, op.namespace, op.key)
                    results.append(item)
                elif isinstance(op, SearchOp):
                    candidates = await self._filter_items(db, op)
                    search_ops[i] = (op, candidates)
                    results.append(None)
                elif isinstance(op, ListNamespacesOp):
                    namespaces = await self._list_namespaces(db, op)
                    results.append(namespaces)
                elif isinstance(op, PutOp):
                    put_ops[(op.namespace, op.key)] = op
                    results.append(None)
                else:
                    raise ValueError(f"Unknown operation type: {type(op)}")

            if search_ops:
                query_vectors = await self._embed_search_queries(search_ops)
                await self._batch_search(db, search_ops, query_vectors, results)

            to_embed = self._extract_texts(put_ops)
            if to_embed and self.index_config and self.embeddings:
                embeddings = await self.embeddings.aembed_documents(list(to_embed))
                await self._insert_vectors(db, to_embed, embeddings)

            await self._apply_put_ops(db, put_ops)
            await db.commit()

            return results

    async def _get_item(
        self, db: aiosqlite.Connection, namespace: Tuple[str, ...], key: str
    ) -> Optional[Item]:
        """Get an item from the database.

        Args:
            db (aiosqlite.Connection): Database connection
            namespace (Tuple[str, ...]): Item namespace
            key (str): Item key

        Returns:
            Optional[Item]: The item if found, None otherwise
        """
        async with db.execute(
            "SELECT value, created_at, updated_at FROM items WHERE namespace = ? AND key = ?",
            ("/".join(namespace), key)
        ) as cursor:
            row = await cursor.fetchone()
            if row:
                return Item(
                    namespace=namespace,
                    key=key,
                    value=json.loads(row[0]),
                    created_at=datetime.fromisoformat(row[1]),
                    updated_at=datetime.fromisoformat(row[2])
                )
            return None

    async def _filter_items(
        self, db: aiosqlite.Connection, op: SearchOp
    ) -> List[Tuple[Item, List[List[float]]]]:
        """Filter items by namespace and filter function.

        Args:
            db (aiosqlite.Connection): Database connection
            op (SearchOp): Search operation

        Returns:
            List[Tuple[Item, List[List[float]]]]: Filtered items with their vectors
        """
        namespace_prefix = "/".join(op.namespace_prefix)
        query = """
            SELECT namespace, key, value, created_at, updated_at
            FROM items
            WHERE namespace LIKE ?
        """
        params = [f"{namespace_prefix}%"]

        async with db.execute(query, params) as cursor:
            rows = await cursor.fetchall()
            filtered = []
            for row in rows:
                item = Item(
                    namespace=tuple(row[0].split("/")),
                    key=row[1],
                    value=json.loads(row[2]),
                    created_at=datetime.fromisoformat(row[3]),
                    updated_at=datetime.fromisoformat(row[4])
                )
                if not op.filter or all(
                    self._compare_values(item.value.get(key), filter_value)
                    for key, filter_value in op.filter.items()
                ):
                    if op.query and self.index_config:
                        vectors = await self._get_vectors(db, item.namespace, item.key)
                        filtered.append((item, vectors))
                    else:
                        filtered.append((item, []))
            return filtered

    async def _get_vectors(
        self, db: aiosqlite.Connection, namespace: Tuple[str, ...], key: str
    ) -> List[List[float]]:
        """Get vectors for an item.

        Args:
            db (aiosqlite.Connection): Database connection
            namespace (Tuple[str, ...]): Item namespace
            key (str): Item key

        Returns:
            List[List[float]]: List of vectors
        """
        async with db.execute(
            "SELECT vector FROM vectors WHERE namespace = ? AND key = ?",
            ("/".join(namespace), key)
        ) as cursor:
            rows = await cursor.fetchall()
            return [json.loads(row[0]) for row in rows]

    async def _list_namespaces(
        self, db: aiosqlite.Connection, op: ListNamespacesOp
    ) -> List[Tuple[str, ...]]:
        """List namespaces matching the conditions.

        Args:
            db (aiosqlite.Connection): Database connection
            op (ListNamespacesOp): List namespaces operation

        Returns:
            List[Tuple[str, ...]]: List of matching namespaces
        """
        async with db.execute("SELECT DISTINCT namespace FROM items") as cursor:
            rows = await cursor.fetchall()
            namespaces = [tuple(ns.split("/")) for (ns,) in rows]

            if op.match_conditions:
                namespaces = [
                    ns for ns in namespaces
                    if all(self._does_match(condition, ns) for condition in op.match_conditions)
                ]

            if op.max_depth is not None:
                namespaces = sorted({ns[:op.max_depth] for ns in namespaces})
            else:
                namespaces = sorted(namespaces)

            return namespaces[op.offset:op.offset + op.limit]

    async def _embed_search_queries(
        self,
        search_ops: Dict[int, Tuple[SearchOp, List[Tuple[Item, List[List[float]]]]]],
    ) -> Dict[str, List[float]]:
        """Embed search queries.

        Args:
            search_ops (Dict[int, Tuple[SearchOp, List[Tuple[Item, List[List[float]]]]]]): Search operations

        Returns:
            Dict[str, List[float]]: Query embeddings
        """
        query_vectors = {}
        if self.index_config and self.embeddings and search_ops:
            queries = {op.query for (op, _) in search_ops.values() if op.query}
            if queries:
                embeddings = await self.embeddings.aembed_documents(list(queries))
                query_vectors = dict(zip(queries, embeddings))
        return query_vectors

    async def _batch_search(
        self,
        db: aiosqlite.Connection,
        ops: Dict[int, Tuple[SearchOp, List[Tuple[Item, List[List[float]]]]]],
        query_vectors: Dict[str, List[float]],
        results: List[Result],
    ) -> None:
        """Perform batch similarity search.

        Args:
            db (aiosqlite.Connection): Database connection
            ops (Dict[int, Tuple[SearchOp, List[Tuple[Item, List[List[float]]]]]]): Search operations
            query_vectors (Dict[str, List[float]]): Query embeddings
            results (List[Result]): Results list to update
        """
        for i, (op, candidates) in ops.items():
            if not candidates:
                results[i] = []
                continue

            if op.query and query_vectors:
                query_vector = query_vectors[op.query]
                flat_items, flat_vectors = [], []
                scoreless = []

                for item, vectors in candidates:
                    for vector in vectors:
                        flat_items.append(item)
                        flat_vectors.append(vector)
                    if not vectors:
                        scoreless.append(item)

                scores = self._cosine_similarity(query_vector, flat_vectors)
                sorted_results = sorted(
                    zip(scores, flat_items), key=lambda x: x[0], reverse=True
                )

                seen = set()
                kept = []
                for score, item in sorted_results:
                    key = (item.namespace, item.key)
                    if key in seen:
                        continue
                    ix = len(seen)
                    seen.add(key)
                    if ix >= op.offset + op.limit:
                        break
                    if ix < op.offset:
                        continue
                    kept.append((score, item))

                if scoreless and len(kept) < op.limit:
                    kept.extend(
                        (None, item) for item in scoreless[:op.limit - len(kept)]
                    )

                results[i] = [
                    SearchItem(
                        namespace=item.namespace,
                        key=item.key,
                        value=item.value,
                        created_at=item.created_at,
                        updated_at=item.updated_at,
                        score=float(score) if score is not None else None,
                    )
                    for score, item in kept
                ]
            else:
                results[i] = [
                    SearchItem(
                        namespace=item.namespace,
                        key=item.key,
                        value=item.value,
                        created_at=item.created_at,
                        updated_at=item.updated_at,
                    )
                    for (item, _) in candidates[op.offset:op.offset + op.limit]
                ]

    async def _apply_put_ops(
        self, db: aiosqlite.Connection, put_ops: Dict[Tuple[Tuple[str, ...], str], PutOp]
    ) -> None:
        """Apply put operations to the database.

        Args:
            db (aiosqlite.Connection): Database connection
            put_ops (Dict[Tuple[Tuple[str, ...], str], PutOp]): Put operations
        """
        for (namespace, key), op in put_ops.items():
            if op.value is None:
                await db.execute(
                    "DELETE FROM items WHERE namespace = ? AND key = ?",
                    ("/".join(namespace), key)
                )
            else:
                now = datetime.now(timezone.utc)
                await db.execute(
                    """
                    INSERT INTO items (namespace, key, value, created_at, updated_at)
                    VALUES (?, ?, ?, ?, ?)
                    ON CONFLICT (namespace, key) DO UPDATE SET
                        value = excluded.value,
                        updated_at = excluded.updated_at
                    """,
                    (
                        "/".join(namespace),
                        key,
                        json.dumps(op.value),
                        now.isoformat(),
                        now.isoformat(),
                    )
                )

    async def _insert_vectors(
        self,
        db: aiosqlite.Connection,
        to_embed: Dict[str, List[Tuple[Tuple[str, ...], str, str]]],
        embeddings: List[List[float]],
    ) -> None:
        """Insert vector embeddings into the database.

        Args:
            db (aiosqlite.Connection): Database connection
            to_embed (Dict[str, List[Tuple[Tuple[str, ...], str, str]]]): Texts to embed
            embeddings (List[List[float]]): Vector embeddings
        """
        indices = [index for indices in to_embed.values() for index in indices]
        if len(indices) != len(embeddings):
            raise ValueError(
                f"Number of embeddings ({len(embeddings)}) does not"
                f" match number of indices ({len(indices)})"
            )

        for embedding, (ns, key, path) in zip(embeddings, indices):
            await db.execute(
                """
                INSERT INTO vectors (namespace, key, path, vector)
                VALUES (?, ?, ?, ?)
                ON CONFLICT (namespace, key, path) DO UPDATE SET
                    vector = excluded.vector
                """,
                ("/".join(ns), key, path, json.dumps(embedding))
            )

    def _extract_texts(
        self, put_ops: Dict[Tuple[Tuple[str, ...], str], PutOp]
    ) -> Dict[str, List[Tuple[Tuple[str, ...], str, str]]]:
        """Extract texts for embedding from put operations.

        Args:
            put_ops (Dict[Tuple[Tuple[str, ...], str], PutOp]): Put operations

        Returns:
            Dict[str, List[Tuple[Tuple[str, ...], str, str]]]: Texts to embed
        """
        if put_ops and self.index_config and self.embeddings:
            to_embed = {}
            for op in put_ops.values():
                if op.value is not None and op.index is not False:
                    if op.index is None:
                        paths = self.index_config["__tokenized_fields"]
                    else:
                        paths = [(ix, tokenize_path(ix)) for ix in op.index]

                    for path, field in paths:
                        texts = get_text_at_path(op.value, field)
                        if texts:
                            if len(texts) > 1:
                                for i, text in enumerate(texts):
                                    to_embed.setdefault(text, []).append(
                                        (op.namespace, op.key, f"{path}.{i}")
                                    )
                            else:
                                to_embed.setdefault(texts[0], []).append(
                                    (op.namespace, op.key, path)
                                )
            return to_embed
        return {}

    def _compare_values(self, item_value: Any, filter_value: Any) -> bool:
        """Compare values in a JSONB-like way.

        Args:
            item_value (Any): Value from the item
            filter_value (Any): Value from the filter

        Returns:
            bool: Whether the values match
        """
        if isinstance(filter_value, dict):
            if any(k.startswith("$") for k in filter_value):
                return all(
                    self._apply_operator(item_value, op_key, op_value)
                    for op_key, op_value in filter_value.items()
                )
            if not isinstance(item_value, dict):
                return False
            return all(
                self._compare_values(item_value.get(k), v)
                for k, v in filter_value.items()
            )
        elif isinstance(filter_value, (list, tuple)):
            return (
                isinstance(item_value, (list, tuple))
                and len(item_value) == len(filter_value)
                and all(
                    self._compare_values(iv, fv)
                    for iv, fv in zip(item_value, filter_value)
                )
            )
        else:
            return item_value == filter_value

    def _apply_operator(self, value: Any, operator: str, op_value: Any) -> bool:
        """Apply a comparison operator.

        Args:
            value (Any): Value to compare
            operator (str): Operator to apply
            op_value (Any): Value to compare against

        Returns:
            bool: Result of the comparison
        """
        if operator == "$eq":
            return value == op_value
        elif operator == "$gt":
            return float(value) > float(op_value)
        elif operator == "$gte":
            return float(value) >= float(op_value)
        elif operator == "$lt":
            return float(value) < float(op_value)
        elif operator == "$lte":
            return float(value) <= float(op_value)
        elif operator == "$ne":
            return value != op_value
        else:
            raise ValueError(f"Unsupported operator: {operator}")

    def _does_match(self, match_condition: MatchCondition, key: Tuple[str, ...]) -> bool:
        """Check if a namespace key matches a match condition.

        Args:
            match_condition (MatchCondition): Match condition to check
            key (Tuple[str, ...]): Namespace key to check

        Returns:
            bool: Whether the key matches the condition
        """
        match_type = match_condition.match_type
        path = match_condition.path

        if len(key) < len(path):
            return False

        if match_type == "prefix":
            for k_elem, p_elem in zip(key, path):
                if p_elem == "*":
                    continue
                if k_elem != p_elem:
                    return False
            return True
        elif match_type == "suffix":
            for k_elem, p_elem in zip(reversed(key), reversed(path)):
                if p_elem == "*":
                    continue
                if k_elem != p_elem:
                    return False
            return True
        else:
            raise ValueError(f"Unsupported match type: {match_type}")

    def _cosine_similarity(self, X: List[float], Y: List[List[float]]) -> List[float]:
        """Compute cosine similarity between a vector X and a matrix Y.

        Args:
            X (List[float]): Query vector
            Y (List[List[float]]): Matrix of vectors to compare against

        Returns:
            List[float]: Cosine similarities
        """
        if not Y:
            return []

        try:
            import numpy as np
            X_arr = np.array(X) if not isinstance(X, np.ndarray) else X
            Y_arr = np.array(Y) if not isinstance(Y, np.ndarray) else Y
            X_norm = np.linalg.norm(X_arr)
            Y_norm = np.linalg.norm(Y_arr, axis=1)

            mask = Y_norm != 0
            similarities = np.zeros_like(Y_norm)
            similarities[mask] = np.dot(Y_arr[mask], X_arr) / (Y_norm[mask] * X_norm)
            return similarities.tolist()
        except ImportError:
            logger.warning(
                "NumPy not found. Using pure Python implementation for vector operations. "
                "This may significantly impact performance. Consider installing NumPy: "
                "pip install numpy"
            )
            similarities = []
            for y in Y:
                dot_product = sum(a * b for a, b in zip(X, y))
                norm1 = sum(a * a for a in X) ** 0.5
                norm2 = sum(a * a for a in y) ** 0.5
                similarity = (
                    dot_product / (norm1 * norm2) if norm1 > 0 and norm2 > 0 else 0.0
                )
                similarities.append(similarity)
            return similarities 


prompt_templates = {
    "review": "You are an expert software engineer with a good taste on how a code should be. Assume that in current working directory, you are in a git repository. Get the current git status and diff. Review the change and provide feedback.",
    "commit": "You are an expert software engineer. Assume that in current working directory, you are in a git repository. Get the current git status and diff. Reason why the change was made. Then commit with a concise, descriptive message that follows Conventional Commits specification.",
    "yt": "Retell and summarize the video in a concise and descriptive manner. Use bullet points and markdown formatting. The url is {url}",
}


from typing import List, Type, Optional, Any, override
from pydantic import BaseModel
from langchain_core.tools import BaseTool, BaseToolkit, ToolException
from mcp import StdioServerParameters, types, ClientSession
from mcp.client.stdio import stdio_client
import pydantic
from pydantic_core import to_json
from jsonschema_pydantic import jsonschema_to_pydantic
import asyncio

from .storage import *

class McpServerConfig(BaseModel):
    """Configuration for an MCP server.
    
    This class represents the configuration needed to connect to and identify an MCP server,
    containing both the server's name and its connection parameters.

    Attributes:
        server_name (str): The name identifier for this MCP server
        server_param (StdioServerParameters): Connection parameters for the server, including
            command, arguments and environment variables
        exclude_tools (list[str]): List of tool names to exclude from this server
    """
    
    server_name: str
    server_param: StdioServerParameters
    exclude_tools: list[str] = []

class McpToolkit(BaseToolkit):
    name: str
    server_param: StdioServerParameters
    exclude_tools: list[str] = []
    _session: Optional[ClientSession] = None
    _tools: List[BaseTool] = []
    _client = None
    _init_lock: asyncio.Lock = None

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, **data):
        super().__init__(**data)
        self._init_lock = asyncio.Lock()

    async def _start_session(self):
        async with self._init_lock:
            if self._session:
                return self._session

            self._client = stdio_client(self.server_param)
            read, write = await self._client.__aenter__()
            self._session = ClientSession(read, write)
            await self._session.__aenter__()
            await self._session.initialize()
            return self._session

    async def initialize(self, force_refresh: bool = False):
        if self._tools and not force_refresh:
            return

        cached_tools = get_cached_tools(self.server_param)
        if cached_tools and not force_refresh:
            for tool in cached_tools:
                if tool.name in self.exclude_tools:
                    continue
                self._tools.append(create_langchain_tool(tool, self._session, self))
            return

        try:
            await self._start_session()
            tools: types.ListToolsResult = await self._session.list_tools()
            save_tools_cache(self.server_param, tools.tools)
            for tool in tools.tools:
                if tool.name in self.exclude_tools:
                    continue
                self._tools.append(create_langchain_tool(tool, self._session, self))
        except Exception as e:
            print(f"Error gathering tools for {self.server_param.command} {' '.join(self.server_param.args)}: {e}")
            raise e
        
    async def close(self):
        try:
            if self._session:
                await self._session.__aexit__(None, None, None)
        except:
            # Currently above code doesn't really works and not closing the session
            # But it's not a big deal as we are exiting anyway
            # TODO find a way to cleanly close the session
            pass
        try:
            if self._client:
                await self._client.__aexit__(None, None, None)
        except:
            # TODO find a way to cleanly close the client
            pass

    def get_tools(self) -> List[BaseTool]:
        return self._tools


class McpTool(BaseTool):
    toolkit_name: str
    name: str
    description: str
    args_schema: Type[BaseModel]
    session: Optional[ClientSession]
    toolkit: McpToolkit

    handle_tool_error: bool = True

    def _run(self, **kwargs):
        raise NotImplementedError("Only async operations are supported")

    async def _arun(self, **kwargs):
        if not self.session:
            self.session = await self.toolkit._start_session()

        result = await self.session.call_tool(self.name, arguments=kwargs)
        content = to_json(result.content).decode()
        if result.isError:
            raise ToolException(content)
        return content

def create_langchain_tool(
    tool_schema: types.Tool,
    session: ClientSession,
    toolkit: McpToolkit,
) -> BaseTool:
    """Create a LangChain tool from MCP tool schema.
    
    Args:
        tool_schema (types.Tool): The MCP tool schema.
        session (ClientSession): The session for the tool.
    
    Returns:
        BaseTool: The created LangChain tool.
    """
    return McpTool(
        name=tool_schema.name,
        description=tool_schema.description,
        args_schema=jsonschema_to_pydantic(tool_schema.inputSchema),
        session=session,
        toolkit=toolkit,
        toolkit_name=toolkit.name,
    )


async def convert_mcp_to_langchain_tools(server_config: McpServerConfig, force_refresh: bool = False) -> McpToolkit:
    """Convert MCP tools to LangChain tools and create a toolkit.
    
    Args:
        server_config (McpServerConfig): Configuration for the MCP server including name and parameters.
        force_refresh (bool, optional): Whether to force refresh the tools cache. Defaults to False.
    
    Returns:
        McpToolkit: A toolkit containing the converted LangChain tools.
    """
    toolkit = McpToolkit(
        name=server_config.server_name, 
        server_param=server_config.server_param,
        exclude_tools=server_config.exclude_tools
    )
    await toolkit.initialize(force_refresh=force_refresh)
    return toolkit



from pathlib import Path

CACHE_EXPIRY_HOURS = 24
DEFAULT_QUERY = "Summarize https://www.youtube.com/watch?v=NExtKbS1Ljc"
CONFIG_FILE = 'mcp-server-config.json'
CONFIG_DIR = Path.home() / ".llm"
SQLITE_DB = CONFIG_DIR / "conversations.db"
CACHE_DIR = CONFIG_DIR / "mcp-tools"


#!/usr/bin/env python3

"""
Simple llm CLI that acts as MCP client.
"""

from datetime import datetime
import argparse
import asyncio
import os
from typing import Annotated, TypedDict
import uuid
import sys
import re
import anyio
from langchain_core.messages import BaseMessage, HumanMessage
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.language_models.chat_models import BaseChatModel
from langgraph.prebuilt import create_react_agent
from langgraph.managed import IsLastStep
from langgraph.graph.message import add_messages
from langchain.chat_models import init_chat_model
from langgraph.checkpoint.sqlite.aio import AsyncSqliteSaver
from rich.console import Console
from rich.table import Table
import base64
import imghdr as imghdr
import mimetypes
from dotenv import load_dotenv, find_dotenv

from .input import *
from .const import *
from .output import *
from .storage import *
from .tool import *
from .prompt import *
from .memory import *
from .config import AppConfig

# The AgentState class is used to maintain the state of the agent during a conversation.
class AgentState(TypedDict):
    # A list of messages exchanged in the conversation.
    messages: Annotated[list[BaseMessage], add_messages]
    # A flag indicating whether the current step is the last step in the conversation.
    is_last_step: IsLastStep
    # The current date and time, used for context in the conversation.
    today_datetime: str
    # The user's memories.
    memories: str = "no memories"
    remaining_steps: int = 5

async def run() -> None:
    """Run the LLM agent."""
    load_dotenv(find_dotenv())
    
    args = setup_argument_parser()
    query, is_conversation_continuation = parse_query(args)
    app_config = AppConfig.load()
    
    if args.list_tools:
        await handle_list_tools(app_config, args)
        return
    
    if args.show_memories:
        await handle_show_memories()
        return
        
    if args.list_prompts:
        handle_list_prompts()
        return
        
    await handle_conversation(args, query, is_conversation_continuation, app_config)

def setup_argument_parser() -> argparse.Namespace:
    """Setup and return the argument parser."""
    parser = argparse.ArgumentParser(
        description='Run LangChain agent with MCP tools',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  llm "What is the capital of France?"     Ask a simple question
  llm c "tell me more"                     Continue previous conversation
  llm p review                             Use a prompt template
  cat file.txt | llm                       Process input from a file
  llm --list-tools                         Show available tools
  llm --list-prompts                       Show available prompt templates
  llm --no-confirmations "search web"      Run tools without confirmation
        """
    )
    parser.add_argument('query', nargs='*', default=[],
                       help='The query to process (default: read from stdin). '
                            'Special prefixes:\n'
                            '  c: Continue previous conversation\n'
                            '  p: Use prompt template')
    parser.add_argument('--list-tools', action='store_true',
                       help='List all available LLM tools')
    parser.add_argument('--list-prompts', action='store_true',
                       help='List all available prompts')
    parser.add_argument('--no-confirmations', action='store_true',
                       help='Bypass tool confirmation requirements')
    parser.add_argument('--force-refresh', action='store_true',
                       help='Force refresh of tools capabilities')
    parser.add_argument('--text-only', action='store_true',
                       help='Print output as raw text instead of parsing markdown')
    parser.add_argument('--no-tools', action='store_true',
                       help='Do not add any tools')
    parser.add_argument('--no-intermediates', action='store_true',
                       help='Only print the final message')
    parser.add_argument('--show-memories', action='store_true',
                       help='Show user memories')
    parser.add_argument('--model',
                       help='Override the model specified in config')
    return parser.parse_args()

async def handle_list_tools(app_config: AppConfig, args: argparse.Namespace) -> None:
    """Handle the --list-tools command."""
    server_configs = [
        McpServerConfig(
            server_name=name,
            server_param=StdioServerParameters(
                command=config.command,
                args=config.args or [],
                env={**(config.env or {}), **os.environ}
            ),
            exclude_tools=config.exclude_tools or []
        )
        for name, config in app_config.get_enabled_servers().items()
    ]
    toolkits, tools = await load_tools(server_configs, args.no_tools, args.force_refresh)
    
    console = Console()
    table = Table(title="Available LLM Tools")
    table.add_column("Toolkit", style="cyan")
    table.add_column("Tool Name", style="cyan")
    table.add_column("Description", style="green")

    for tool in tools:
        if isinstance(tool, McpTool):
            table.add_row(tool.toolkit_name, tool.name, tool.description)

    console.print(table)

    for toolkit in toolkits:
        await toolkit.close()

async def handle_show_memories() -> None:
    """Handle the --show-memories command."""
    store = SqliteStore(SQLITE_DB)
    memories = await get_memories(store)
    console = Console()
    table = Table(title="My LLM Memories")
    for memory in memories:
        table.add_row(memory)
    console.print(table)

def handle_list_prompts() -> None:
    """Handle the --list-prompts command."""
    console = Console()
    table = Table(title="Available Prompt Templates")
    table.add_column("Name", style="cyan")
    table.add_column("Template")
    table.add_column("Arguments")
    
    for name, template in prompt_templates.items():
        table.add_row(name, template, ", ".join(re.findall(r'\{(\w+)\}', template)))
        
    console.print(table)

async def load_tools(server_configs: list[McpServerConfig], no_tools: bool, force_refresh: bool) -> tuple[list, list]:
    """Load and convert MCP tools to LangChain tools."""
    if no_tools:
        return [], []
        
    toolkits = []
    langchain_tools = []
    
    async def convert_toolkit(server_config: McpServerConfig):
        toolkit = await convert_mcp_to_langchain_tools(server_config, force_refresh)
        toolkits.append(toolkit)
        langchain_tools.extend(toolkit.get_tools())

    async with anyio.create_task_group() as tg:
        for server_param in server_configs:
            tg.start_soon(convert_toolkit, server_param)
            
    langchain_tools.append(save_memory)
    return toolkits, langchain_tools

async def handle_conversation(args: argparse.Namespace, query: HumanMessage, 
                            is_conversation_continuation: bool, app_config: AppConfig) -> None:
    """Handle the main conversation flow."""
    server_configs = [
        McpServerConfig(
            server_name=name,
            server_param=StdioServerParameters(
                command=config.command,
                args=config.args or [],
                env={**(config.env or {}), **os.environ}
            ),
            exclude_tools=config.exclude_tools or []
        )
        for name, config in app_config.get_enabled_servers().items()
    ]
    toolkits, tools = await load_tools(server_configs, args.no_tools, args.force_refresh)
    
    extra_body = {}
    if app_config.llm.base_url and "openrouter" in app_config.llm.base_url:
        extra_body = {"transforms": ["middle-out"]}
    # Override model if specified in command line
    if args.model:
        app_config.llm.model = args.model
    
    if app_config.llm.provider == "gigachat":
        from langchain_gigachat import GigaChat
        
        model: BaseChatModel = GigaChat(
            model=app_config.llm.model,
            model_provider=app_config.llm.provider,
            api_key=app_config.llm.api_key,
            temperature=app_config.llm.temperature,
            base_url=app_config.llm.base_url,
            default_headers={
                "X-Title": "mcp-client-cli",
                "HTTP-Referer": "https://github.com/adhikasp/mcp-client-cli",
            },
            extra_body=extra_body
        )
    else:
        model: BaseChatModel = init_chat_model(
            model=app_config.llm.model,
            model_provider=app_config.llm.provider,
            api_key=app_config.llm.api_key,
            temperature=app_config.llm.temperature,
            base_url=app_config.llm.base_url,
            default_headers={
                "X-Title": "mcp-client-cli",
                "HTTP-Referer": "https://github.com/adhikasp/mcp-client-cli",
            },
            extra_body=extra_body
        )

    prompt = ChatPromptTemplate.from_messages([
        ("system", app_config.system_prompt),
        ("placeholder", "{messages}")
    ])

    conversation_manager = ConversationManager(SQLITE_DB)
    
    async with AsyncSqliteSaver.from_conn_string(SQLITE_DB) as checkpointer:
        store = SqliteStore(SQLITE_DB)
        memories = await get_memories(store)
        formatted_memories = "\n".join(f"- {memory}" for memory in memories)
        agent_executor = create_react_agent(
            model, tools, state_schema=AgentState, 
            state_modifier=prompt, checkpointer=checkpointer, store=store
        )
        
        thread_id = (await conversation_manager.get_last_id() if is_conversation_continuation 
                    else uuid.uuid4().hex)

        input_messages = AgentState(
            messages=[query], 
            today_datetime=datetime.now().isoformat(),
            memories=formatted_memories,
            remaining_steps=3
        )

        output = OutputHandler(text_only=args.text_only, only_last_message=args.no_intermediates)
        output.start()
        try:
            async for chunk in agent_executor.astream(
                input_messages,
                stream_mode=["messages", "values"],
                config={"configurable": {"thread_id": thread_id, "user_id": "myself"}, 
                       "recursion_limit": 100}
            ):
                output.update(chunk)
                if not args.no_confirmations:
                    if not output.confirm_tool_call(app_config.__dict__, chunk):
                        break
        except Exception as e:
            output.update_error(e)
        finally:
            output.finish()

        await conversation_manager.save_id(thread_id, checkpointer.conn)

    for toolkit in toolkits:
        await toolkit.close()

def parse_query(args: argparse.Namespace) -> tuple[HumanMessage, bool]:
    """
    Parse the query from command line arguments.
    Returns a tuple of (HumanMessage, is_conversation_continuation).
    """
    query_parts = ' '.join(args.query).split()
    stdin_content = ""
    stdin_image = None
    is_continuation = False

    # Handle clipboard content if requested
    if query_parts and query_parts[0] == 'cb':
        # Remove 'cb' from query parts
        query_parts = query_parts[1:]
        # Try to get content from clipboard
        clipboard_result = get_clipboard_content()
        if clipboard_result:
            content, mime_type = clipboard_result
            if mime_type:  # It's an image
                stdin_image = base64.b64encode(content).decode('utf-8')
            else:  # It's text
                stdin_content = content
        else:
            print("No content found in clipboard")
            raise Exception("Clipboard is empty")
    # Check if there's input from pipe
    elif not sys.stdin.isatty():
        stdin_data = sys.stdin.buffer.read()
        # Try to detect if it's an image
        image_type = imghdr.what(None, h=stdin_data)
        if image_type:
            # It's an image, encode it as base64
            stdin_image = base64.b64encode(stdin_data).decode('utf-8')
            mime_type = mimetypes.guess_type(f"dummy.{image_type}")[0] or f"image/{image_type}"
        else:
            # It's text
            stdin_content = stdin_data.decode('utf-8').strip()

    # Process the query text
    query_text = ""
    if query_parts:
        if query_parts[0] == 'c':
            is_continuation = True
            query_text = ' '.join(query_parts[1:])
        elif query_parts[0] == 'p' and len(query_parts) >= 2:
            template_name = query_parts[1]
            if template_name not in prompt_templates:
                print(f"Error: Prompt template '{template_name}' not found.")
                print("Available templates:", ", ".join(prompt_templates.keys()))
                return HumanMessage(content=""), False

            template = prompt_templates[template_name]
            template_args = query_parts[2:]
            try:
                # Extract variable names from the template
                var_names = re.findall(r'\{(\w+)\}', template)
                # Create dict mapping parameter names to arguments
                template_vars = dict(zip(var_names, template_args))
                query_text = template.format(**template_vars)
            except KeyError as e:
                print(f"Error: Missing argument {e}")
                return HumanMessage(content=""), False
        else:
            query_text = ' '.join(query_parts)

    # Combine stdin content with query text if both exist
    if stdin_content and query_text:
        query_text = f"{stdin_content}\n\n{query_text}"
    elif stdin_content:
        query_text = stdin_content
    elif not query_text and not stdin_image:
        return HumanMessage(content=""), False

    # Create the message content
    if stdin_image:
        content = [
            {"type": "text", "text": query_text or "What do you see in this image?"},
            {"type": "image_url", "image_url": {"url": f"data:{mime_type};base64,{stdin_image}"}}
        ]
    else:
        content = query_text

    return HumanMessage(content=content), is_continuation

def main() -> None:
    """Entry point of the script."""
    asyncio.run(run())


if __name__ == "__main__":
    main()



"""Configuration management for the MCP client CLI."""

from dataclasses import dataclass
from pathlib import Path
import os
import commentjson
from typing import Dict, List, Optional

from .const import CONFIG_FILE, CONFIG_DIR

@dataclass
class LLMConfig:
    """Configuration for the LLM model."""
    model: str = "gpt-4o"
    provider: str = "openai"
    api_key: Optional[str] = None
    temperature: float = 0
    base_url: Optional[str] = None

    @classmethod
    def from_dict(cls, config: dict) -> "LLMConfig":
        """Create LLMConfig from dictionary."""
        return cls(
            model=config.get("model", cls.model),
            provider=config.get("provider", cls.provider),
            api_key=config.get("api_key", os.getenv("LLM_API_KEY", os.getenv("OPENAI_API_KEY", ""))),
            temperature=config.get("temperature", cls.temperature),
            base_url=config.get("base_url"),
        )

@dataclass
class ServerConfig:
    """Configuration for an MCP server."""
    command: str
    args: List[str] = None
    env: Dict[str, str] = None
    enabled: bool = True
    exclude_tools: List[str] = None
    requires_confirmation: List[str] = None

    @classmethod
    def from_dict(cls, config: dict) -> "ServerConfig":
        """Create ServerConfig from dictionary."""
        return cls(
            command=config["command"],
            args=config.get("args", []),
            env=config.get("env", {}),
            enabled=config.get("enabled", True),
            exclude_tools=config.get("exclude_tools", []),
            requires_confirmation=config.get("requires_confirmation", [])
        )

@dataclass
class AppConfig:
    """Main application configuration."""
    llm: LLMConfig
    system_prompt: str
    mcp_servers: Dict[str, ServerConfig]
    tools_requires_confirmation: List[str]

    @classmethod
    def load(cls) -> "AppConfig":
        """Load configuration from file."""
        config_paths = [CONFIG_FILE, CONFIG_DIR / "config.json"]
        chosen_path = next((path for path in config_paths if os.path.exists(path)), None)
        
        if chosen_path is None:
            raise FileNotFoundError(f"Could not find config file in any of: {', '.join(map(str, config_paths))}")

        with open(chosen_path, 'r') as f:
            config = commentjson.load(f)

        # Extract tools requiring confirmation
        tools_requires_confirmation = []
        for server_config in config["mcpServers"].values():
            tools_requires_confirmation.extend(server_config.get("requires_confirmation", []))

        return cls(
            llm=LLMConfig.from_dict(config.get("llm", {})),
            system_prompt=config["systemPrompt"],
            mcp_servers={
                name: ServerConfig.from_dict(server_config)
                for name, server_config in config["mcpServers"].items()
            },
            tools_requires_confirmation=tools_requires_confirmation
        )

    def get_enabled_servers(self) -> Dict[str, ServerConfig]:
        """Get only enabled server configurations."""
        return {
            name: config 
            for name, config in self.mcp_servers.items() 
            if config.enabled
        } 


from mcp.server.fastmcp import FastMCP
from pydantic import BaseModel
import sys

mcp = FastMCP("Math")

class Query(BaseModel):
    query: str


class Person(BaseModel):
    name: str
    age: int


@mcp.tool()
def find_preson(name: Query) -> Person:
    """Find an info about some person by name"""
    print(f">>> Calling find_person({name})")
    return Person(name="John Doe", age=30)


@mcp.tool()
def add(a: float, b: float) -> float:
    """Add two numbers"""
    print(f">>> Calling add({a}, {b})")
    return a + b


@mcp.tool()
def multiply(a: float, b: float) -> float:
    """Multiply two numbers"""
    print(f">>> Calling multiply({a}, {b})")
    return a * b


if __name__ == "__main__":
    transport = sys.argv[1] if len(sys.argv) > 1 else "stdio"
    mcp.run(transport=transport)



from datetime import datetime, timedelta
from typing import Optional, List
from mcp import StdioServerParameters, types
import json
import aiosqlite
import uuid

from .const import *

def get_cached_tools(server_param: StdioServerParameters) -> Optional[List[types.Tool]]:
    """Retrieve cached tools if available and not expired.
    
    Args:
        server_param (StdioServerParameters): The server parameters to identify the cache.
    
    Returns:
        Optional[List[types.Tool]]: A list of tools if cache is available and not expired, otherwise None.
    """
    CACHE_DIR.mkdir(parents=True, exist_ok=True)
    cache_key = f"{server_param.command}-{'-'.join(server_param.args)}".replace("/", "-")
    cache_file = CACHE_DIR / f"{cache_key}.json"
    
    if not cache_file.exists():
        return None
        
    cache_data = json.loads(cache_file.read_text())
    cached_time = datetime.fromisoformat(cache_data["cached_at"])
    
    if datetime.now() - cached_time > timedelta(hours=CACHE_EXPIRY_HOURS):
        return None
            
    return [types.Tool(**tool) for tool in cache_data["tools"]]


def save_tools_cache(server_param: StdioServerParameters, tools: List[types.Tool]) -> None:
    """Save tools to cache.
    
    Args:
        server_param (StdioServerParameters): The server parameters to identify the cache.
        tools (List[types.Tool]): The list of tools to be cached.
    """
    cache_key = f"{server_param.command}-{'-'.join(server_param.args)}".replace("/", "-")
    cache_file = CACHE_DIR / f"{cache_key}.json"
    
    cache_data = {
        "cached_at": datetime.now().isoformat(),
        "tools": [tool.model_dump() for tool in tools]
    }
    cache_file.write_text(json.dumps(cache_data))


class ConversationManager:
    """Manages conversation persistence in SQLite database."""
    
    def __init__(self, db_path: Path):
        self.db_path = db_path
        self.db_path.parent.mkdir(parents=True, exist_ok=True)
    
    async def _init_db(self, db) -> None:
        """Initialize database schema.
        
        Args:
            db: The database connection object.
        """
        await db.execute("""
            CREATE TABLE IF NOT EXISTS last_conversation (
                id INTEGER PRIMARY KEY,
                thread_id TEXT NOT NULL
            )
        """)
        await db.commit()
    
    async def get_last_id(self) -> str:
        """Get the thread ID of the last conversation.
        
        Returns:
            str: The thread ID of the last conversation, or a new UUID if no conversation exists.
        """
        async with aiosqlite.connect(self.db_path) as db:
            await self._init_db(db)
            async with db.execute("SELECT thread_id FROM last_conversation LIMIT 1") as cursor:
                row = await cursor.fetchone()
            return row[0] if row else uuid.uuid4().hex
    
    async def save_id(self, thread_id: str, db = None) -> None:
        """Save thread ID as the last conversation.
        
        Args:
            thread_id (str): The thread ID to save.
            db: The database connection object (optional).
        """
        if db is None:
            async with aiosqlite.connect(self.db_path) as db:
                await self._save_id(db, thread_id)
        else:
            await self._save_id(db, thread_id)
    
    async def _save_id(self, db, thread_id: str) -> None:
        """Internal method to save thread ID.
        
        Args:
            db: The database connection object.
            thread_id (str): The thread ID to save.
        """
        async with db.cursor() as cursor:
            await self._init_db(db)
            await cursor.execute("DELETE FROM last_conversation")
            await cursor.execute(
                "INSERT INTO last_conversation (thread_id) VALUES (?)", 
                (thread_id,)
            )
            await db.commit()